# -*- coding: utf-8 -*-
"""
Created on Sun Jul 25 13:47:37 2021

@author: Ngan H. Nguyen

Intruction: This code is specific to Social Influence Model; however, it is easy to turn it into Axelrod Model 
by modifying the probability metric in function neighborChange(self, row, col). In order to speed up the code, 
many thresholds were used and were specific to a certain set of paramaters. Those who wish to replicate this 
project should start with the suggestive paramaters (size=4, listLength= 5, traits=10, rows=30, cols=30) in 
this code and slowly increase or decrease some or all parameters to determine (guess) these threshold. However,
if your machine is powerfull and you are very patient, feel free to eliminate the threshold. In addition, when 
one particular parameter (such as number of frozen zones) is targeted, it be more reasonable to comment out all 
other outputs to speed up the process. Some of the figures (Figure 2, 14) in the assoicated published paper can 
also be easily obtains by slightly modify the probability metric according to the given formula in the paper. 
For Figures 3 and 4, instead of using createRandom(self), one just needs to assign a particular matrix for 
self.data and write a few modified linse to get the desired outputs. 
"""

import random
import time
import statistics
import numpy as np
import matplotlib.pyplot as plt
import math

#Grid Class

class Grid:

   #initializes Grid
   def __init__(self, size, listLength, traits, rows, cols):
       self.width = cols
       self.height = rows
       self.length = listLength # number of features
       self.traits = traits
       self.size = size # neighborhood's size
       W = self.width
       H = self.height
       L = self.length
       self.data = [[['0' for leng in range(L) ] for widt in range(W)] for row in range(H) ]
       self.neighbors = [[[] for widt in range(W)] for row in range(H) ]

   #allows grid to be printed
   def __repr__(self):
       H = self.height
       W = self.width
       L = self.length
       s = ''   # the string to return
       for row in range(0,H):        
           for col in range(0,W):
               s += '|'
               for leng in range(0,L):
                   s+= self.data[row][col][leng]
           s += '\n'
       return s

   #creates a random board           
   def createRandom(self):
       H = self.height
       W = self.width
       L = self.length
       T = self.traits
       S = self.size
       K = []
       for row in range(0,H):
           for col in range(0,W):
               self.neighbors[row][col] = self.neighborhood(row,col,S)
               for leng in range(0,L):
                   for i in range(T):
                       K.append(str(i))
                       self.data[row][col][leng]=random.choice(K)
   
   # defines a neighborhood
                    
   def neighborhood(self,row,col,size):
       H=self.height
       #W=self.width 
       neighborhood=[]
       if size == 4:
           neighborhood.append([row+1,col])
           neighborhood.append([row-1,col])
           neighborhood.append([row,col+1])
           neighborhood.append([row,col-1])
       elif size == 8:
           neighborhood.append([row+1,col])
           neighborhood.append([row-1,col])
           neighborhood.append([row,col+1])
           neighborhood.append([row,col-1])
           neighborhood.append([row+1,col+1])
           neighborhood.append([row-1,col+1])
           neighborhood.append([row+1,col-1])
           neighborhood.append([row-1,col-1])
       elif size == 12:
           neighborhood.append([row+1,col])
           neighborhood.append([row-1,col])
           neighborhood.append([row,col+1])
           neighborhood.append([row,col-1])
           neighborhood.append([row+1,col+1])
           neighborhood.append([row-1,col+1])
           neighborhood.append([row+1,col-1])
           neighborhood.append([row-1,col-1])
           neighborhood.append([row+2,col])
           neighborhood.append([row-2,col])
           neighborhood.append([row,col+2])
           neighborhood.append([row,col-2])
       # re-defines neighborhood on the grid's borders
       for j in range(2):
           neighborhood = [i for i in neighborhood if i[j]<H and i[j]>=0]
       """ 
       # Neighborhood on the torus
       if size == 4:
           neighborhood.append([(row+1)%H,col])
           neighborhood.append([(row-1)%H,col])
           neighborhood.append([row,(col+1)%W])
           neighborhood.append([row,(col-1)%W])
       elif size == 8:
           neighborhood.append([(row+1)%H,col])
           neighborhood.append([(row-1)%H,col])
           neighborhood.append([row,(col+1)%W])
           neighborhood.append([row,(col-1)%W])
           neighborhood.append([(row+1)%H,(col+1)%W])
           neighborhood.append([(row-1)%H,(col+1)%W])
           neighborhood.append([(row+1)%H,(col-1)%W])
           neighborhood.append([(row-1)%H,(col-1)%W])
       elif size == 12:
           neighborhood.append([(row+1)%H,col])
           neighborhood.append([(row-1)%H,col])
           neighborhood.append([row,(col+1)%W])
           neighborhood.append([row,(col-1)%W])
           neighborhood.append([(row+1)%H,(col+1)%W])
           neighborhood.append([(row-1)%H,(col+1)%W])
           neighborhood.append([(row+1)%H,(col-1)%W])
           neighborhood.append([(row-1)%H,(col-1)%W])
           neighborhood.append([(row+2)%H,col])
           neighborhood.append([(row-2)%H,col])
           neighborhood.append([row,(col+2)%W])
           neighborhood.append([row,(col-2)%W])       
               
       """    
       return neighborhood


   #counts the number of regions 
   def countRegion(self):
       S = self.size
       arr = self.data
       x = len(arr) #height
       y = len(arr[0]) #width
       #create a new array the same size as the original
       regions = [[None for i in range(y)] for i in range(x)]
       #print(regions)
       label = 0
       queue = []
       def check_neighbor(i, j, v):
           if not regions[i][j] and arr[i][j] == v:
               regions[i][j] = label
               queue.insert(0, (i, j))
       for i in range(x):
           for j in range(y):
               #don't check an already labelled region
               if regions[i][j]: continue
               label += 1 #new label
               regions[i][j] = label
               queue = [(i, j)]
               v = arr[i][j]
               #keep checking neighbours until we run out
               if S == 4:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
               elif S == 8:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)   
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
                       if X > 0 and Y > 0:
                           check_neighbor(X-1, Y-1, v)
                       if X < x-1 and Y > 0:
                           check_neighbor(X+1, Y-1, v)
                       if X < x-1 and Y < y-1:
                           check_neighbor(X+1, Y+1, v)
                       if X > 0 and Y < y-1:
                           check_neighbor(X-1, Y+1, v)
               elif S == 12:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 1:
                           check_neighbor(X-2, Y, v)
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-2:
                           check_neighbor(X+2, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)
                       if Y > 1:
                           check_neighbor(X, Y-2, v)    
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-2:
                           check_neighbor(X, Y+2, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
                       if X > 0 and Y > 0:
                           check_neighbor(X-1, Y-1, v)
                       if X < x-1 and Y > 0:
                           check_neighbor(X+1, Y-1, v)
                       if X < x-1 and Y < y-1:
                           check_neighbor(X+1, Y+1, v)
                       if X > 0 and Y < y-1:
                           check_neighbor(X-1, Y+1, v)
       return label # this is the number of regions

   #Returns a list of different traits between two agents
   def differentTraits(self, row1, col1, neighbor):
       DifferentTraits = []
       L=self.length
       for i in range(0,L):
           if self.data[row1][col1][i] != self.data[neighbor[0]][neighbor[1]][i]:
               DifferentTraits.append(i)
       return DifferentTraits

   #Returns a random integer representing a row
   def getRandomRow(self):
       H = self.height
       row = random.randint(0,H-1)
       return row

   #Returns a random interger representing a column
   def getRandomCol(self):
       W = self.width
       col = random.randint(0,W-1)
       return col

   #Returns true or false if the bond is active
   def activeBond(self, row1, col1, neighbor):
       L = self.length
       DifferentTraits=self.differentTraits(row1, col1, neighbor)
       if len(DifferentTraits)!=0 and len(DifferentTraits) != L:
           n=random.randint(0,L-1)
           if self.data[row1][col1][n] == self.data[neighbor[0]][neighbor[1]][n]:
               return True

   #Returns true or false if there is a possible mov ein the grid  
   def possibleMove(self):
       L = self.length
       H = self.height
       W = self.width
       S = self.size
       # compares row-agents
       for row in range(0,H-1):
           for col in range(0,W):
               if S == 4:
                   if self.data[row][col]!=self.data[row+1][col]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row+1][col][n]:
                               return True
               elif S == 8:
                   if self.data[row][col]!=self.data[row+1][col]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row+1][col][n]:
                               return True
                   if col!=W-1:
                       if self.data[row][col]!=self.data[row+1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row+1][col+1][n]:
                                   return True
               elif S==12:
                   if self.data[row][col]!=self.data[row+1][col]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row+1][col][n]:
                               return True
                   if row!=H-2:
                       if self.data[row][col]!=self.data[row+2][col]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row+2][col][n]:
                                   return True
                   if col!=W-1:
                       if self.data[row][col]!=self.data[row+1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row+1][col+1][n]:
                                   return True
       # compares column-agents
       for row in range(0,H):
           for col in range(0,W-1):
               if S == 4:
                   if self.data[row][col]!=self.data[row][col+1]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row][col+1][n]:
                               return True
               elif S == 8:
                   if self.data[row][col]!=self.data[row][col+1]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row][col+1][n]:
                               return True
                   if row!=0:
                       if self.data[row][col]!=self.data[row-1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row-1][col+1][n]:
                                   return True
               elif S==12:
                   if self.data[row][col]!=self.data[row][col+1]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row][col+1][n]:
                               return True
                   if col!=W-2:
                       if self.data[row][col]!=self.data[row][col+2]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row][col+2][n]:
                                   return True
                   if row!=0:
                       if self.data[row][col]!=self.data[row-1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row-1][col+1][n]:
                                   return True                  
       return False
   
   #Returns an integer that is the number of features shared
   def overlap(self, row, col, neighbors, feature):
       overlap=0
       for i in neighbors:
           if self.data[row][col][feature]==self.data[i[0]][i[1]][feature]:
               overlap+=1
       return overlap
  
   def regionSize(self):
       arr = self.data
       Ss = self.size
       x = len(arr) #height
       y = len(arr[0]) #width
       R = []
       sizes = []
       #create a new array the same size as the original
       regions = [[None for i in range(y)] for i in range(x)]
       #print(regions)
       label = 0
       queue = []
       def check_neighbor(i, j, v):
           if not regions[i][j] and arr[i][j] == v: #if neighbor is None
               regions[i][j] = label
               queue.insert(0, (i, j)) #insert neighbor to the front of the list queue
               return 1
           else:
               return 0
       for i in range(x):
           for j in range(y):
               #don't check an already labelled region
               if regions[i][j]:
                   continue
               R.append(arr[i][j]) # new region be added
               S = 1 #the size of the new region
               label += 1 #new label
               regions[i][j] = label
               queue = [(i, j)]
               v = arr[i][j]
               #keep checking neighbours until we run out
               if Ss == 4:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
               elif Ss == 8:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)   
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
                       if X > 0 and Y > 0:
                           check_neighbor(X-1, Y-1, v)
                       if X < x-1 and Y > 0:
                           check_neighbor(X+1, Y-1, v)
                       if X < x-1 and Y < y-1:
                           check_neighbor(X+1, Y+1, v)
                       if X > 0 and Y < y-1:
                           check_neighbor(X-1, Y+1, v)
               elif Ss == 12:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 1:
                           check_neighbor(X-2, Y, v)
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-2:
                           check_neighbor(X+2, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)
                       if Y > 1:
                           check_neighbor(X, Y-2, v)    
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-2:
                           check_neighbor(X, Y+2, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
                       if X > 0 and Y > 0:
                           check_neighbor(X-1, Y-1, v)
                       if X < x-1 and Y > 0:
                           check_neighbor(X+1, Y-1, v)
                       if X < x-1 and Y < y-1:
                           check_neighbor(X+1, Y+1, v)
                       if X > 0 and Y < y-1:
                           check_neighbor(X-1, Y+1, v)
               sizes.append(S)
       return sizes



   #Goes through the process of changing a neighbor
   def neighborChange(self, row, col):
       neighborhood=self.neighbors[row][col]
       #Random Neighbor Chosen
       neighbor = random.choice(neighborhood)  
       #Calculate probability of interaction
       if self.activeBond(row, col, neighbor)==True:
           #Interaction occurs so we proceed
           DifferentTraits = self.differentTraits(row, col, neighbor)        
           #Choose random unshared feature between i and j
           feature = random.choice(DifferentTraits)
           #calculate overlap between neighborhood and agent and selected neighbor
           agentOverlap = self.overlap(row, col, neighborhood, feature)
           neighborOverlap = self.overlap(neighbor[0], neighbor[1], neighborhood, feature)
           #add one as the agent compares itself to itself
           agentOverlap += 1
           #Calculate P2
           P2 = neighborOverlap/(agentOverlap+neighborOverlap)
           if random.random() < P2:
               self.data[row][col][feature] = self.data[neighbor[0]][neighbor[1]][feature]
               return 1
       return 0
   
   # stripped down Flache rule 
   def neighborChangeFlache(self, row, col):
       neighborhood=self.neighbors[row][col]
       #Random Neighbor Chosen
       Influences = []
       unsharedFeatures = []
       for neighbor in neighborhood:      
           #Calculate probability of interaction
           if self.activeBond(row, col, neighbor)==True:
               #Interaction occurs so we proceed
               for f in self.differentTraits(row, col, neighbor):
                   if f not in unsharedFeatures:
                       unsharedFeatures.append(f)
               Influences.append(neighbor)
       if unsharedFeatures!=[]:
           exchangeFeature = random.choice(unsharedFeatures)
           vote = {}
           for neighbor in Influences:
                if self.data[neighbor[0]][neighbor[1]][exchangeFeature] in vote:
                       vote[self.data[neighbor[0]][neighbor[1]][exchangeFeature]]+=1
                else:
                       vote[self.data[neighbor[0]][neighbor[1]][exchangeFeature]] =1
           maxvote = max(vote.values())
           modal = []
           for i in vote:
               if vote[i] == maxvote:
                   modal.append(i)
           if self.data[row][col][exchangeFeature] in modal:
               return 0
           else:
                self.data[row][col][exchangeFeature] = random.choice(modal)
                return 1         
       return 0
   
   def disorder1(self):
       H = self.height
       W = self.width
       L = self.length
       LikeFeatures=0
       NumConnections=0
       totalcomparisons=0
       for row in range(0,H-1):
           for col in range(0,W):
               NumConnections+=1
               for n in range(0,L):
                   totalcomparisons+=1
                   if self.data[row][col][n]== self.data[row+1][col][n]:
                       LikeFeatures+=1
       for row in range(0,H):
           for col in range(0,W-1):
               NumConnections+=1
               for n in range(0,L):
                   totalcomparisons+=1
                   if self.data[row][col][n] == self.data[row][col+1][n]:
                       LikeFeatures+=1
       unlikeFeatures=totalcomparisons-LikeFeatures
       disorder=unlikeFeatures/(NumConnections*L)
       return disorder
   
   def disorder2(self):
       Sizes=self.regionSize0()
       H=self.height
       Sizes.sort()
       Sizes.insert(0,0)
       x=0
       N=H*H
       M=len(Sizes)
       for i in range(1,M):
           N+= -Sizes[i-1]
           p= math.sqrt(N)+1-math.sqrt(Sizes[i])
           x += 2*math.log(p)
       y=math.log(math.factorial(H*H))
       disorder = x/y
       return disorder
   
   def disorder3(self):
       Ss = self.size
       arr = self.data
       x = len(arr) #height
       y = len(arr[0]) #width
       #create a new array the same size as the original
       regions = [[None for i in range(y)] for i in range(x)]
       #print(regions)
       label = 0
       queue = []
       def check_neighbor(i, j, v):
           if not regions[i][j] and arr[i][j] == v:
               regions[i][j] = label
               queue.insert(0, (i, j))
       for i in range(x):
           for j in range(y):
               #don't check an already labelled region
               if regions[i][j]: continue
               label += 1 #new label
               regions[i][j] = label
               queue = [(i, j)]
               v = arr[i][j]
               #keep checking neighbours until we run out
               if Ss == 4:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
               elif Ss == 8:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)   
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
                       if X > 0 and Y > 0:
                           check_neighbor(X-1, Y-1, v)
                       if X < x-1 and Y > 0:
                           check_neighbor(X+1, Y-1, v)
                       if X < x-1 and Y < y-1:
                           check_neighbor(X+1, Y+1, v)
                       if X > 0 and Y < y-1:
                           check_neighbor(X-1, Y+1, v)
               elif Ss == 12:
                   while queue:
                       (X, Y) = queue.pop()
                       if X > 1:
                           check_neighbor(X-2, Y, v)
                       if X > 0:
                           check_neighbor(X-1, Y, v)
                       if X < x-2:
                           check_neighbor(X+2, Y, v)
                       if X < x-1:
                           check_neighbor(X+1, Y, v)
                       if Y > 1:
                           check_neighbor(X, Y-2, v)    
                       if Y > 0:
                           check_neighbor(X, Y-1, v)
                       if Y < y-2:
                           check_neighbor(X, Y+2, v)
                       if Y < y-1:
                           check_neighbor(X, Y+1, v)
                       if X > 0 and Y > 0:
                           check_neighbor(X-1, Y-1, v)
                       if X < x-1 and Y > 0:
                           check_neighbor(X+1, Y-1, v)
                       if X < x-1 and Y < y-1:
                           check_neighbor(X+1, Y+1, v)
                       if X > 0 and Y < y-1:
                           check_neighbor(X-1, Y+1, v)
       stuff=[[] for i in range(label)]
       for i in range(x):
           for j in range(y):
               stuff[regions[i][j]-1].append([i,j])
       connections=[0 for i in range(label)]
       for i in range(label):
           for j in stuff[i]:
               hood = self.neighbors[j[0]][j[1]]
               for n in hood:
                   if self.data[j[0]][j[1]] == self.data[n[0]][n[1]]:
                       connections[i] += 1
       connections = [i/2 for i in connections]
       regionSizes=[0 for i in range(label)]
       for i in range(label):
           regionSizes[i] = len(stuff[i])
       borderlengths=[0 for i in range(label)]
       for i in range(label):
           borderlengths[i] = 4*regionSizes[i]-2*connections[i]
       N = x**2
       M = len(regionSizes)
       if M==1 or M == N:
           return 0      
       disorder = 0
       for i in range(label):
           disorder = disorder + borderlengths[i]/(4*math.sqrt(regionSizes[i]))
       disorder = disorder/M - 1
       return disorder

   #Gathers statistics
   def loop(self,n):
       H=self.height
       W=self.width
       q=0
       regions=[]
       Disorder1=[]
       Disorder2=[]
       Disorder3=[]
       times=[]
       matrices=[]
       interactions=[]    
       events=[]
       sizes=[]
       #a loop of n
       while q!=n:
           plt.close('all')
           self.createRandom()
           event=0
           X=[]
           Y=[]
           Z1=[]
           Z2=[]
           Z3=[]
           success=0
           name = str(q)+ 'data' +'.txt'
           f = open(name, "w")        
           fname = str(q)+'chart' +'.png'
           disorder1= str(q)+'disorder1'+'.png'
           disorder2= str(q)+'disorder2'+'.png'
           disorder3= str(q)+'disorder3'+'.png'
           imageName = str(q)+'map'+'.png'
           regionName = str(q)+'regionsize'+'.png'
           f.write('Starting grid')
           f.write('\n')
           for item in self.data:
               f.write(str(item))
               if item!=self.data[H-1][W-1]:
                   f.write(',')
           f.write('\n')
           f.write('\n')       
           f.write('Run: ')
           f.write(str(q))
           f.write('\n')
           start = time.time()
           a=0
           b=0
           #a process
           while a==0:
               row = self.getRandomRow()
               col = self.getRandomCol()           
               b = self.neighborChange(row,col)
               success += b
               if event==0 or success%100==0: # the number here is for the suggestive parameters set mentioned in the Introduction
                   if event==0 or b!=0:
                       countRegion=self.countRegion()
                       X.append(event) # an alternative why is to use successful interaction instead of event by X.append(success)
                       Y.append(countRegion) 
                       Z1.append(self.disorder1())
                       Z2.append(self.disorder2())
                       Z3.append(self.disorder3())
                       
               #out put every 60000 success. The numbers here is for the suggestive parameters set mentioned in the Introduction
               if event==0 or success in [20000,40000,60000,100000,180000,300000,450000]: 
                   if event==0 or b!=0:
                       borderName = str(q)+'Runs'+str(event)+'border'+'.png'
                       self.color1(borderName)
                       plt.close()
                       f.write('events: ')
                       f.write(str(event))
                       f.write('\n')
                       f.write('Successful interactions: ')
                       f.write(str(success))
                       f.write('\n')
                       f.write('Number of regions: ')
                       f.write(str(countRegion))
                       f.write('\n')
                       f.write('\n')
               event+=1 
               if success>=250000 and b!=0: # the numbers here is for the suggestive parameters set mentioned in the Introduction
                   if self.possibleMove()==False:
                       a=1       
           events.append(event) #or X.append(success)
           Z1.append(self.disorder1())
           Z2.append(self.disorder2())
           Z3.append(self.disorder3())
           
           #Massive list of matrices
           Matrix=[[[] for j in range(W)] for i in range(H)]
           for i in range (H):
               for j in range (W):
                   for x in self.data[i][j]:
                       Matrix[i][j].append(str(x))
           matrices.append([Matrix]) 
           borderName = str(q)+'Runs9999999finalBorder'+'.png'
           end = time.time()
           totalTime = end - start
           regions.append(self.countRegion())
           times.append(totalTime)
           interactions.append(success)
           plt.ylim(ymax=self.width)
           plt.plot(X,Y, linewidth=1)
           plt.savefig(fname)
           plt.close('all')
           X.append(event)
           plt.ylim(0,1)
           plt.plot(X,Z1, linewidth=1)
           plt.savefig(disorder1)
           plt.close('all')
           plt.ylim(0,1)
           plt.plot(X,Z2, linewidth=1)
           plt.savefig(disorder2)
           plt.close('all')
           plt.ylim(0,1) # this max is often sufficient but not always
           plt.plot(X,Z3, linewidth=1)
           plt.savefig(disorder3)
           plt.close('all')  
           self.color(imageName)
           self.color1(borderName)
           s=self.regionSizes()
           for i in range (len(s)):
               sizes.append(s[i])
           plt.close('all')
              
           Disorder1.append(self.disorder1())
           Disorder2.append(self.disorder2())
           Disorder3.append(self.disorder3())
           
           #Output final state(after a completed process )
           f.write('\n')
           f.write('\n')
           f.write('Final State')
           for item in self.data:
               f.write(str(item))
               if item!=self.data[H-1][W-1]:
                   f.write(',')
           f.write('\n')
           f.write('\n')
           f.write("Regions: ")
           f.write(str(self.countRegion()))
           f.write('\n')
           f.write('\n')
           f.write('Total Time: ')
           f.write(str(totalTime))
           f.write('\n')
           f.write('\n')
           f.write('Total events: ')
           f.write(str(event))
           f.write('\n')
           f.write('\n')
           f.write('Total successful interactions: ')
           f.write(str(success))
           f.write('\n')
           f.write('\n')
           f.write('Disorder1: ')
           f.write(str(self.disorder1()))
           f.write('\n')
           f.write('\n')
           f.write('Disorder2: ')
           f.write(str(self.disorder2()))
           f.write('\n')
           f.write('\n')
           f.write('Disorder3: ')
           f.write(str(self.disorder3()))
           f.write('\n')
           f.write('\n')
           f.close()
           q+=1
          
       #Out put after a completed loop
       avgRegion = statistics.mean(regions)
       avgTime = statistics.mean(times)
       avgInteractions = statistics.mean(interactions)
       avgEvents = statistics.mean(events)
       avgDisorder1= statistics.mean(Disorder1)
       avgDisorder2= statistics.mean(Disorder2)
       avgDisorder3= statistics.mean(Disorder3)
       
       
       # generates a histogram for the suggestive parameters set mentioned in the Introduction. Please use the appropriate limits and steps for different parameter.
       O=[]
       K=[]
       sizes.sort()
       for i in sizes:
           if i not in K:
               K.append(i)
       for j in K:
           O.append(sizes.count(j))
       for i in a:
           K.append(i)
           O.append(a[i]) 
       plt.figure(figsize=(20,40))
       plt.ylim(0,400)
       plt.bar(K,O, edgecolor='k',color='b')
       plt.xticks(np.arange(0, 901, step=100),size=40,fontname="Times New Roman")
       plt.yticks(size=40,fontname="Times New Roman")
       plt.title('Frozen Zone Sizes versus Number of Occurrences \n in the Social Influence Model',fontweight="bold",fontname="Times New Roman",size=55)
       plt.xlabel('Frozen Zone Sizes',fontname="Times New Roman", size=50)
       plt.ylabel('Number of Occurences',fontname="Times New Roman",size=50)
       plt.savefig(regionName,bbox_inches='tight')
       plt.show()	
       plt.close()
       
       
       E=0
       for i in range (n):
           e=((avgRegion-regions[i])**2)/(n)
           E+=e
       g = open('stats.txt', "w")
       print(regions)
       print('Average stable regions number: ',avgRegion)
       print('Standard error: ',math.sqrt(E)/math.sqrt(n))
       print('Average time: ',avgTime)
       g.write('Average stable regions number: ')
       g.write(str(avgRegion))
       g.write('\n')
       g.write('Stand Error: ')
       g.write(str(math.sqrt(E)/math.sqrt(n)))
       g.write('\n')
       g.write('Average Time: ')
       g.write(str(avgTime))
       g.write('\n')
       g.write('Average Interactions: ')
       g.write(str(avgInteractions))
       g.write('\n')
       g.write('Average Events: ')
       g.write(str(avgEvents))
       g.write('\n')
       g.write('Average Disorder1: ')
       g.write(str(avgDisorder1))
       g.write('\n')
       g.write('\n')
       g.write('Average Disorder2: ')
       g.write(str(avgDisorder2))
       g.write('\n')
       g.write('\n')
       g.write('Average Disorder3: ')
       g.write(str(avgDisorder3))
       g.write('\n')
       g.write('\n')
       
       # saves all the grids
       for i in matrices:
           for j in i:
               g.write(str(j))
               if item!=i[0][H-1][W-1]:
                   g.write(',')
       g.close()

   #Returns a colored grid
   def color(self,imageName):
       H = self.height
       W = self.width
       Matrix = np.zeros((H, W))
       stableRegions=[]
       for row in range (0,H):
           for col in range(0,W):
               if self.data[row][col] not in stableRegions:
                   stableRegions.append(self.data[row][col])
       for row in range (0,H):
           for col in range(0,W):
               for i in range (len(stableRegions)):
                   if self.data[row][col]==stableRegions[i]:
                       Matrix[row,col]=i+1   
       Matrix=np.flip(Matrix,0)
       xi = np.arange(0, W+1)
       yi = np.arange(0, H+1)
       X, Y = np.meshgrid(xi, yi)
       plt.pcolormesh(X, Y, Matrix)
       plt.pcolormesh(Matrix, edgecolors="w")
       plt.axis('off')
       plt.title('Stable State of Social Influence Model')
       plt.savefig(imageName)
       plt.close()

   def color1(self, borderName):
       H = self.height #NumY
       W = self.width  #NumX
       L = self.length
       l=500/(W*H)
       for row in range(0,H):
           for col in range(0,W-1):
               m=0
               for n in range(0,L):
                   if self.data[row][col][n]==self.data[row][col+1][n]:
                       m+=1
               x=[W-row,W-row-1]
               y=[col+1,col+1]
               if m/L==1:
                   plt.plot(y,x,zorder=1,c='white',linewidth=l)
               elif m/L>=.8:
                   plt.plot(y,x,zorder=1,c='lightblue',linewidth=l)
               elif m/L>=.6:
                   plt.plot(y,x,zorder=1,c='lightskyblue',linewidth=l)
               elif m/L>=.4:
                   plt.plot(y,x,zorder=1,c='deepskyblue',linewidth=l)
               elif m/L>=.2:
                   plt.plot(y,x,zorder=1,c='blue',linewidth=l)
               elif m/L>0:
                   plt.plot(y,x,zorder=1,c='midnightblue',linewidth=l)
               else:
                   plt.plot(y,x,zorder=1,c='black',linewidth=l)
       for row in range(0,H-1):
           for col in range(0,W):
               m=0
               for n in range(0,L):
                   if self.data[row][col][n]==self.data[row+1][col][n]:
                       m+=1
               x=[W-row-1,W-row-1]
               y=[col,col+1]
               if m/L==1:
                   plt.plot(y,x,zorder=1,c='white',linewidth=l)
               elif m/L>=.8:
                   plt.plot(y,x,zorder=1,c='lightblue',linewidth=l)
               elif m/L>=.6:
                   plt.plot(y,x,zorder=1,c='lightskyblue',linewidth=l)
               elif m/L>=.4:
                   plt.plot(y,x,zorder=1,c='deepskyblue',linewidth=l)
               elif m/L>=.2:
                   plt.plot(y,x,zorder=1,c='blue',linewidth=l)
               elif m/L>0:
                   plt.plot(y,x,zorder=1,c='midnightblue',linewidth=l)
               else:
                   plt.plot(y,x,zorder=1,c='black',linewidth=l)            
       for row in range (H+1):
           for col in range (W+1):
               plt.scatter(row, col,s=l*l, c='k', marker="o", zorder=2)
       plt.title('Evolving State of Social Influence Model')
       plt.axis('off')
       #plt.show()
       plt.savefig(borderName)
       plt.close()

   """
   #Faster code for large value of parameter. A lot of output lines were still kept 
   #but they are should be commented out, especially those which are collected every few steps
   
   def neighborChange(self, link):
        L = self.length       
        #Interaction occurs so we proceed
        r = random.randrange(0,L)
        if self.data[link[0]][link[1]][r] == self.data[link[2]][link[3]][r]:
            DifferentTraits = []
            for i in range(0,L):
               if self.data[link[0]][link[1]][i] != self.data[link[2]][link[3]][i]:
                   DifferentTraits.append(i)      
            #Choose random unshared feature between i and j        
            neighborhood=self.neighbors[link[0]][link[1]]
            feature = random.choice(DifferentTraits)
            agentOverlap = self.overlap(link[0], link[1], neighborhood, feature)
            neighborOverlap = self.overlap(link[2], link[3], neighborhood, feature)
            P2 = neighborOverlap/(agentOverlap+neighborOverlap)
            if random.random() < P2:
                self.data[link[0]][link[1]][feature] = self.data[link[2]][link[3]][feature]
                for j in self.neighbors[link[0]][link[1]]:
                    if self.data[link[0]][link[1]] != self.data[j[0]][j[1]]:
                        temp = 0
                        for n in range(0,L):
                            if self.data[link[0]][link[1]][n] == self.data[j[0]][j[1]][n]:
                                temp +=1
                                break                                               
                        if temp == 0:
                            if [link[0],link[1],j[0],j[1]]  in self.link:
                                self.link=[i for i in self.link if i!=[link[0],link[1],j[0],j[1]]]
                            elif [j[0],j[1],link[0],link[1]] in self.link :
                                self.link=[i for i in self.link if i!=[j[0],j[1],link[0],link[1]]]
                        else:
                            if ([link[0],link[1],j[0],j[1]] not in self.link) and ([j[0],j[1],link[0],link[1]] not in self.link) :
                                self.link=self.link+[[link[0],link[1],j[0],j[1]]]
                    else:
                        if [link[0],link[1],j[0],j[1]]  in self.link:                           
                            self.link=[i for i in self.link if i!=[link[0],link[1],j[0],j[1]]]                       
                        elif [j[0],j[1],link[0],link[1]] in self.link :                       
                            self.link=[i for i in self.link if i!=[j[0],j[1],link[0],link[1]]]                               
                return 1
        return 0
   def track(self): 
       S=self.size
       self.link = []
       H = self.height
       W = self.width
       L = self.length
       # compares row-agents
       for row in range(0,H-1):
           for col in range(0,W):
               if S == 4:                  
                   if self.data[row][col]!=self.data[row+1][col]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row+1][col][n]:
                               self.link.append([row,col,row+1,col])
                               break
               elif S == 8:
                   if self.data[row][col]!=self.data[row+1][col]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row+1][col][n]:
                               self.link.append([row,col,row+1,col])
                               break
                   if col!=W-1:
                       if self.data[row][col]!=self.data[row+1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row+1][col+1][n]:
                                   self.link.append([row,col,row+1,col+1])
                                   break
               elif S==12:
                   if self.data[row][col]!=self.data[row+1][col]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row+1][col][n]:
                               self.link.append([row,col,row+1,col])
                               break
                   if row!=H-2:
                       if self.data[row][col]!=self.data[row+2][col]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row+2][col][n]:
                                   self.link.append([row,col,row+2,col])
                                   break
                   if col!=W-1:
                       if self.data[row][col]!=self.data[row+1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row+1][col+1][n]:
                                   self.link.append([row,col,row+1,col+1])
                                   break
       # compares column-agents
       for row in range(0,H):
           for col in range(0,W-1):
               if S == 4:         
                   if self.data[row][col]!=self.data[row][col+1]:
                        for n in range(0,L):
                            if self.data[row][col][n]==self.data[row][col+1][n]:
                                 self.link.append([row,col,row,col+1])
                                 break
               elif S == 8:
                   if self.data[row][col]!=self.data[row][col+1]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row][col+1][n]:
                               self.link.append([row,col,row,col+1])
                               break
                   if row!=0:
                       if self.data[row][col]!=self.data[row-1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row-1][col+1][n]:
                                   self.link.append([row,col,row-1,col+1])
                                   break
               elif S==12:
                   if self.data[row][col]!=self.data[row][col+1]:
                       for n in range(0,L):
                           if self.data[row][col][n]==self.data[row][col+1][n]:
                               self.link.append([row,col,row,col+1])
                               break
                   if col!=W-2:
                       if self.data[row][col]!=self.data[row][col+2]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row][col+2][n]:
                                   self.link.append([row,col,row,col+2])
                                   break
                   if row!=0:
                       if self.data[row][col]!=self.data[row-1][col+1]:
                           for n in range(0,L):
                               if self.data[row][col][n]==self.data[row-1][col+1][n]:
                                   self.link.append([row,col,row-1,col+1])
                                   break          
       return self.link 
   
   def loop(self,n):
       H=self.height
       W=self.width
       q=0
       regions=[]
       Disorder1=[]
       Disorder2=[]
       Disorder3=[]
       times=[]
       matrices=[]
       interactions=[]    
       sizes=[]
       #a loop of n
       while q!=n:
           plt.close('all')
           self.createRandom()
           X=[]
           Y=[]
           Z1=[]
           Z2=[]
           Z3=[]
           success=0
           name = str(q)+ 'data' +'.txt'
           f = open(name, "w")        
           fname = str(q)+'chart' +'.png'
           disorder1= str(q)+'disorder1'+'.png'
           disorder2= str(q)+'disorder2'+'.png'
           disorder3= str(q)+'disorder3'+'.png'
           imageName = str(q)+'map'+'.png'
           regionName = str(q)+'regionsize'+'.png'
           f.write('Starting grid')
           f.write('\n')
           for item in self.data:
               f.write(str(item))
               if item!=self.data[H-1][W-1]:
                   f.write(',')
           f.write('\n')
           f.write('\n')       
           f.write('Run: ')
           f.write(str(q))
           f.write('\n')
           start = time.time()
           a=0
           b=0        
           #the number here is suggestive. Each parameter set has different number.
           #These numbers should be when the system have the number of zone equals to a third of its total agents.
           while success < 400000 and time.time() - start < 40: 
                 row = self.getRandomRow()
                 col = self.getRandomCol()     
                 b = self.neighborChange1(row, col)
                 success+=b 
                 if success%100==0: # the number here is for the suggestive parameters set mentioned in the Introduction
                    if  b!=0:
                        countRegion=self.countRegion()
                        X.append(success)
                        Y.append(countRegion) 
                        Z1.append(self.disorder1())
                        Z2.append(self.disorder2())
                        Z3.append(self.disorder3())
                       
                 #out put every 60000 success. The numbers here is for the suggestive parameters set mentioned in the Introduction
                 if success in [20000,40000,60000,100000,180000,300000,450000]: 
                    if b!=0:
                       borderName = str(q)+'Runs'+str(success)+'border'+'.png'
                       self.color1(borderName)
                       plt.close()
                       f.write('Successful interactions: ')
                       f.write(str(success))
                       f.write('\n')
                       f.write('Number of regions: ')
                       f.write(str(countRegion))
                       f.write('\n')
                       f.write('\n')
           links = self.track()
           #print(success)
           #print(len(links))
           while self.link != []:
                link = random.choice(self.link)
                if random.randrange(4)>1:
                    link=[link[2],link[3],link[0],link[1]]
                b = self.neighborChange(link)
                success+=b 
                if success%100==0: # the number here is for the suggestive parameters set mentioned in the Introduction
                    if  b!=0:
                        countRegion=self.countRegion()
                        X.append(success)
                        Y.append(countRegion) 
                        Z1.append(self.disorder1())
                        Z2.append(self.disorder2())
                        Z3.append(self.disorder3())
                       
                #out put every 60000 success. The numbers here is for the suggestive parameters set mentioned in the Introduction
                if success in [20000,40000,60000,100000,180000,300000,450000]: 
                    if b!=0:
                       borderName = str(q)+'Runs'+str(success)+'border'+'.png'
                       self.color1(borderName)
                       plt.close()
                       f.write('Successful interactions: ')
                       f.write(str(success))
                       f.write('\n')
                       f.write('Number of regions: ')
                       f.write(str(countRegion))
                       f.write('\n')
                       f.write('\n')

           Z1.append(self.disorder1())
           Z2.append(self.disorder2())
           Z3.append(self.disorder3())
           X.append(success)
           #Massive list of matrices
           Matrix=[[[] for j in range(W)] for i in range(H)]
           for i in range (H):
               for j in range (W):
                   for x in self.data[i][j]:
                       Matrix[i][j].append(str(x))
           matrices.append([Matrix]) 
           borderName = str(q)+'Runs9999999finalBorder'+'.png'
           end = time.time()
           totalTime = end - start
           regions.append(self.countRegion())
           times.append(totalTime)
           interactions.append(success)
           plt.ylim(ymax=self.width)
           plt.plot(X,Y, linewidth=1)
           plt.savefig(fname)
           plt.close('all')
           X.append(success)
           plt.ylim(0,1)
           plt.plot(X,Z1, linewidth=1)
           plt.savefig(disorder1)
           plt.close('all')
           plt.ylim(0,1)
           plt.plot(X,Z2, linewidth=1)
           plt.savefig(disorder2)
           plt.close('all')
           plt.ylim(0,1) # this max is often sufficient but not always
           plt.plot(X,Z3, linewidth=1)
           plt.savefig(disorder3)
           plt.close('all')  
           self.color(imageName)
           self.color1(borderName)
           s=self.regionSizes()
           for i in range (len(s)):
               sizes.append(s[i])
           plt.close('all')
              
           Disorder1.append(self.disorder1())
           Disorder2.append(self.disorder2())
           Disorder3.append(self.disorder3())
           
           #Output final state(after a completed process )
           f.write('\n')
           f.write('\n')
           f.write('Final State')
           for item in self.data:
               f.write(str(item))
               if item!=self.data[H-1][W-1]:
                   f.write(',')
           f.write('\n')
           f.write('\n')
           f.write("Regions: ")
           f.write(str(self.countRegion()))
           f.write('\n')
           f.write('\n')
           f.write('Total Time: ')
           f.write(str(totalTime))
           f.write('\n')
           f.write('\n')
           f.write('Total successful interactions: ')
           f.write(str(success))
           f.write('\n')
           f.write('\n')
           f.write('Disorder1: ')
           f.write(str(self.disorder1()))
           f.write('\n')
           f.write('\n')
           f.write('Disorder2: ')
           f.write(str(self.disorder2()))
           f.write('\n')
           f.write('\n')
           f.write('Disorder3: ')
           f.write(str(self.disorder3()))
           f.write('\n')
           f.write('\n')
           f.close()
           q+=1
          
       #Out put after a completed loop
       avgRegion = statistics.mean(regions)
       avgTime = statistics.mean(times)
       avgInteractions = statistics.mean(interactions)
       avgDisorder1= statistics.mean(Disorder1)
       avgDisorder2= statistics.mean(Disorder2)
       avgDisorder3= statistics.mean(Disorder3)
       
       
       # generates a histogram for the suggestive parameters set mentioned in the Introduction. Please use the appropriate limits and steps for different parameter.
       O=[]
       K=[]
       sizes.sort()
       for i in sizes:
           if i not in K:
               K.append(i)
       for j in K:
           O.append(sizes.count(j))
       for i in a:
           K.append(i)
           O.append(a[i]) 
       plt.figure(figsize=(20,40))
       plt.ylim(0,400)
       plt.bar(K,O, edgecolor='k',color='b')
       plt.xticks(np.arange(0, 901, step=100),size=40,fontname="Times New Roman")
       plt.yticks(size=40,fontname="Times New Roman")
       plt.title('Frozen Zone Sizes versus Number of Occurrences \n in the Social Influence Model',fontweight="bold",fontname="Times New Roman",size=55)
       plt.xlabel('Frozen Zone Sizes',fontname="Times New Roman", size=50)
       plt.ylabel('Number of Occurences',fontname="Times New Roman",size=50)
       plt.savefig(regionName,bbox_inches='tight')
       plt.show()	
       plt.close()
       
       
       E=0
       for i in range (n):
           e=((avgRegion-regions[i])**2)/(n)
           E+=e
       g = open('stats.txt', "w")
       print(regions)
       print('Average stable regions number: ',avgRegion)
       print('Standard error: ',math.sqrt(E)/math.sqrt(n))
       print('Average time: ',avgTime)
       g.write('Average stable regions number: ')
       g.write(str(avgRegion))
       g.write('\n')
       g.write('Stand Error: ')
       g.write(str(math.sqrt(E)/math.sqrt(n)))
       g.write('\n')
       g.write('Average Time: ')
       g.write(str(avgTime))
       g.write('\n')
       g.write('Average Interactions: ')
       g.write(str(avgInteractions))
       g.write('\n')
       g.write('Average Disorder1: ')
       g.write(str(avgDisorder1))
       g.write('\n')
       g.write('\n')
       g.write('Average Disorder2: ')
       g.write(str(avgDisorder2))
       g.write('\n')
       g.write('\n')
       g.write('Average Disorder3: ')
       g.write(str(avgDisorder3))
       g.write('\n')
       g.write('\n')
       
       # saves all the grids
       for i in matrices:
           for j in i:
               g.write(str(j))
               if item!=i[0][H-1][W-1]:
                   g.write(',')
       g.close()
   """
